#!/usr/bin/env python
#

# Copyright (c) 2016-2019, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import signal
from string import Template
from uuid import uuid4
from base64 import urlsafe_b64encode
from copy import deepcopy
from collections import defaultdict
import subprocess
import shutil
import time
import pkgutil
import pkg_resources
import socket
import resource
import datetime
import sys


class Cluster (object):
    def __init__(self, name, root_path, nodes=['localhost'], debug=False,
                 cleanup=True):
        super(Cluster, self).__init__()
        self.debug = debug
        self.do_cleanup = cleanup
        self.name = name
        # Generate a uuid for this cluster, may be used as see fit by Apps.
        # The uuid is Base64-encoded (without padding) to match the
        # Kafka randomUuid format.
        self.uuid = urlsafe_b64encode(uuid4().bytes).decode('utf-8').strip('=')
        self.instance = str(int(time.time()))[2:]
        self.nodes = dict()
        for n in nodes:
            self.nodes[n] = Node(n)

        if sys.platform == 'linux2':
            self.platform = 'linux'
        else:
            self.platform = sys.platform

        self.apps = list()

        self.root_path = os.path.join(os.path.abspath(root_path), name)

        self.appid_next = 1
        # Allocated TcpPortAllocator ports
        self.tcp_ports = dict()

    def log(self, msg):
        print('[%s] %s: %s' % (datetime.datetime.now(), self.name, msg))

    def dbg(self, msg):
        if self.debug:
            return self.log(msg)

    def find_node(self, nodename):
        return self.nodes.get(nodename, None)

    def get_node(self):
        """ Returns a node """
        for node in self.nodes:
            return self.nodes[node]

    def add_app(self, app):
        self.apps.append(app)

    def find_app(self, appclass, by_conf=None):
        """ Return an app instance matching appclass (string or type).
            If by_conf is set to a (name,value) tuple, the application's
            config property 'name' must have the value of 'value'.
        """
        for app in self.apps:
            if type(appclass) == str:
                if app.__class__.__name__ != appclass:
                    continue
            elif not isinstance(app, appclass):
                continue

            if by_conf is not None:
                if app.conf.get(by_conf[0], None) != by_conf[1]:
                    continue
            return app

        return None

    def find_apps(self, appclass, in_state=None):
        """ Returns a list of app instances matching appclass (type). """
        apps = []
        for app in self.apps:
            if not isinstance(app, appclass):
                continue

            if in_state is not None and app.state != in_state:
                continue

            apps.append(app)

        return apps

    def deploy(self):
        """ @brief Deploy all apps in cluster """
        for app in self.apps:
            app.deploy()

    def start(self, timeout=None):
        """
        Start all apps in cluster
        @param timeout float/None: Number of seconds to wait for cluster to
                                   go operational, raises an
                                   Exception on failure.
        """
        for app in self.apps:
            if app.autostart and app.status() != 'started':
                app.start()

        if timeout is not None and not self.wait_operational(timeout):
            raise Exception('Cluster did not go operational in %ds' % timeout)

    def stop(self, force=False):
        """ Stop all apps in cluster """
        for app in reversed(self.apps):
            app.stop(force=force)

    def cleanup(self, keeptypes=['perm', 'log']):
        if not self.do_cleanup:
            return
        for app in reversed(self.apps):
            app.cleanup(keeptypes=keeptypes)

    def run_post_cmds(self):
        """
        Run any registered post_start_cmds for all apps.
        Should only be called once when the cluster goes fully operational.
        """
        for app in self.apps:
            app.run_post_cmds()

    def wait_stopped(self, timeout=30):
        """ Wait for all apps to stop """
        t_end = time.time() + timeout
        while time.time() < t_end:
            not_stopped = [x for x in self.apps if x.status() != 'stopped']
            if len(not_stopped) == 0:
                return True
            self.dbg('Waiting for %d apps to stop: %s' %
                     (len(not_stopped),
                      ', '.join([str(x) for x in not_stopped])))
            time.sleep(1)
        return False

    def wait_operational(self, timeout=30):
        """ Wait for all started apps in the cluster to become operational """
        t_end = time.time() + timeout
        while time.time() < t_end:
            not_oper = [x for x in self.apps
                        if x.status() == 'started' and not x.operational()]
            stopped = [x for x in self.apps if x.status() == 'stopped']
            if len(not_oper) == 0:
                if len(stopped) > 0:
                    self.log(('%d apps terminated while waiting '
                              'to go operational: %s') %
                             (len(stopped),
                              ', '.join([str(x) for x in stopped])))
                    return False
                # Run post_start_cmds for all apps
                self.run_post_cmds()
                return True
            self.dbg('Waiting for %d apps to go operational: %s' %
                     (len(not_oper), ', '.join([str(x) for x in not_oper])))
            time.sleep(1)
        return False

    def get_all(self, key, defval=None, match_class=None):
        """ Retrieve key from all apps, return as list. """
        return [x.get(key, defval) for x in self.apps
                if isinstance(x, match_class)]

    def instance_path(self):
        """ Returns the instance path """
        return os.path.join(self.root_path, self.instance)

    def mkpath(self, relpath, unique=False, in_instance=False):
        """ Cluster-wide path: will not be cleaned up. """
        if in_instance:
            path = os.path.join(self.root_path, self.instance, relpath)
        else:
            path = os.path.join(self.root_path, relpath)
        if unique is True:
            path += '.' + str(uuid4())
        return path

    def get_docker_host(self):
        """ Get the docker host's name or IP from a container's point of view.
            On OSX we use the handy host.docker.internal hostname, on Linux
            we assume the docker container runs with --network=host and thus
            has access to localhost ports. """
        if self.platform == 'darwin':
            return 'host.docker.internal'
        elif self.platform == 'linux':
            return 'localhost'

        raise Exception("Unsupported platform: {}".format(self.platform))


class Allocator (object):
    def __init__(self, cluster):
        super(Allocator, self).__init__()
        self.cluster = cluster

    def next(self, app=None):
        appid = self.cluster.appid_next
        self.cluster.appid_next += 1
        return appid


class TcpPortAllocator (Allocator):
    def next(self, app, port_base=None):
        """ Let the kernel allocate a port number by opening a TCP socket,
            then closing it and return the port number.
            Linux tries to avoid returning the same port again, so this should
            work...
        """
        if port_base is not None:
            port = port_base
        else:
            port = 0

        for i in range(1, 100):
            s = None
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.bind(('', port))
                port = s.getsockname()[1]
                s.close()
            except Exception:
                if s is not None:
                    s.close()
                if port_base is not None:
                    port += 1
                    continue
                raise
            if self.cluster.tcp_ports.get(port, None) is not None:
                port += 1
                continue
            self.cluster.tcp_ports[port] = app
            return port

        raise Exception(("Could not allocate port (port_base=%s) "
                         "in 100 attempts") % port_base)


class UuidAllocator (Allocator):
    @staticmethod
    def _next(trunc=36):
        return str(uuid4())[:trunc]

    def next(self, app, trunc=36):
        return self._next(trunc=trunc)


class Node (object):
    def __init__(self, name):
        super(Node, self).__init__()
        self.name = name
        if name == 'localhost':
            self.exec_cmd = ''
        else:
            self.exec_cmd = 'ssh %s ' % name


class App (object):
    def __init__(self, cluster, conf=None, on=None):
        self.appid = Allocator(cluster).next(self)
        self.name = self.__class__.__name__
        self.cluster = cluster
        self.autostart = True  # Starts with cluster.start()
        self.do_cleanup = True
        # Environment variables applied to execution
        self.env = defaultdict(list)
        self.env_add('LC_ALL', 'C')
        # List {'type': .., 'path': ..} tuples of created paths, for cleanup
        self.paths = list()
        self.debug = cluster.debug
        # Post-startup commands
        self.post_start_cmds = list()

        self.t_started = 0
        self.t_stopped = 0

        if on:
            self.node = cluster.find_node(on)
        else:
            self.node = cluster.get_node()
        if self.node is None:
            raise Exception('No node available for %s' % self)

        cluster.add_app(self)
        if conf is None:
            self.conf = dict()
        else:
            self.conf = deepcopy({k:v for k, v in conf.items() if v is not None})

        self.conf['appid'] = self.appid
        self.conf['name'] = self.name
        self.conf['nodename'] = self.node.name

        if 'version' not in self.conf:
            self.conf['version'] = 'master'

        # Runtime root path (runtime created files)
        self._root_path = os.path.join(cluster.root_path, cluster.instance,
                                       self.name)

        # Create root path dir
        self.create_dir('')

        self.state = 'init'
        self.dbg('Creating %s instance' % self.name)

    def log(self, msg):
        print('[%s] %s-%s: %s' %
              (datetime.datetime.now(), self.name, self.appid, msg))

    def dbg(self, msg):
        if self.debug:
            return self.log(msg)

    def get(self, key, defval=None):
        """ Return conf value for @param key, or @param defval if not found."""
        return self.conf.get(key, defval)

    def root_path(self):
        return os.path.join(self._root_path, str(self.appid))

    def add_path(self, relpath, pathtype):
        """ Add path for future use by cleanup() et.al. """
        self.paths.append({'path': relpath, 'type': pathtype})
        return relpath

    def mkpath(self, relpath, pathtype='temp', unique=False):
        """ pathtype := perm, temp, log """
        path = os.path.join(self.root_path(), relpath)
        if unique is True:
            path += '.' + str(uuid4())
        self.add_path(path, pathtype)
        return path

    def create_dir(self, relpath, unique=False):
        path = self.mkpath(relpath, unique=unique)
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    def open_file(self, relpath, unique=False, pathtype='temp'):
        path = self.mkpath(relpath, unique=unique, pathtype=pathtype)
        basename = os.path.dirname(path)
        if not os.path.exists(basename):
            os.makedirs(basename)
        f = open(path, 'wb')

        return f, path

    def create_file(self, relpath, unique=False, data=None, pathtype='temp'):
        f, path = self.open_file(relpath, unique=unique, pathtype=pathtype)
        if data is not None:
            if type(data) == str:
                data = data.encode('ascii')
            f.write(data)
        f.close()
        return path

    def create_file_from_template(self, relpath, unique=False,
                                  template_name=None, append_data=None,
                                  subst=True, pathtype='temp'):
        """ Create file from app template using app's conf dict.
            If subst=False no template operations will be performed
            and the file is copied verbatim. """
        if not template_name:
            tname = template_name = os.path.basename(relpath)
        else:
            tname = template_name

        # Try pkgutil resource locator
        tpath = os.path.join('apps', self.__class__.__name__,
                             tname + '.template')
        filedata = pkgutil.get_data('trivup', tpath)
        if filedata is None:
            raise IOError('Class %s resource %s not found' % ('trivup', tpath))

        if subst:
            rendered = Template(filedata.decode('ascii')).substitute(self.conf)
        else:
            rendered = filedata.decode('ascii')
        if append_data is not None:
            rendered += '\n' + append_data + '\n'
        return self.create_file(relpath, unique, data=rendered,
                                pathtype=pathtype)

    def resource_path(self, relpath):
        """ @returns the full path to an application class resource file """
        return pkg_resources.resource_filename('trivup',
                                               os.path.join('apps', self.__class__.__name__, relpath))  # noqa: E501

    def env_add(self, name, value, append=True):
        """ Add (overwrite or append) environment variable """
        if name in self.env and append:
            self.env[name] += ' %s' % value
        else:
            self.env[name] = value

    def start_cmd(self):
        """ @return Command line to start application. """
        return self.conf['start_cmd']

    def execute(self, cmd, stdout_fd=None, stderr_fd=None):
        """
        Execute command, returns the subprocess handle

        @param stdout_fd, stderr_fd: either None (for no redirect), a fd,
                                     or a string (to open and append to file)
        """
        cmd = self.node.exec_cmd + cmd
        self.dbg('Executing: %s' % cmd)
        self.dbg('Environment: %s' % str(self.env))

        fdlimit = self.conf.get('fdlimit', 0)
        if fdlimit == 'max':
            try:
                soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
                fdlimit = hard
            except Exception as e:
                self.log(('Failed to get RLIMIT_NOFILE: {}: '
                          'using system default').format(e))
                fdlimit = 0
        self.dbg('FD limit: %d' % fdlimit)
        if fdlimit > 0:
            try:
                resource.setrlimit(resource.RLIMIT_NOFILE, (fdlimit, fdlimit))
            except ValueError as e:
                self.log('Failed to set RLIMIT_NOFILE({},{}): {}'.format(
                    fdlimit, fdlimit, e))

        to_close = list()
        if type(stdout_fd) == str:
            f = open(stdout_fd, 'a')
            stdout_fd = f.fileno()
            to_close.append(f)
        if type(stderr_fd) == str:
            f = open(stderr_fd, 'a')
            stderr_fd = f.fileno()
            to_close.append(f)

        f = open('/dev/null', 'r')
        stdin_fd = f.fileno()
        to_close.append(f)

        proc = subprocess.Popen(cmd, shell=True, preexec_fn=os.setsid,
                                env=dict(os.environ, **self.env),
                                stdout=stdout_fd, stderr=stderr_fd,
                                stdin=stdin_fd)
        for f in to_close:
            f.close()

        return proc

    def run(self):
        """ Run application using conf @param start_cmd.
            @returns true if app was started, or false if no start_cmd """
        start_cmd = self.start_cmd()
        if start_cmd is None:
            return False

        self.stdout_fd = open(self.mkpath('stdout.log', pathtype='log'), 'a')
        self.stderr_fd = open(self.mkpath('stderr.log', pathtype='log'), 'a')
        self.proc = self.execute(start_cmd,
                                 stdout_fd=self.stdout_fd,
                                 stderr_fd=self.stderr_fd)
        return True

    def run_post_cmds(self):
        """
        Run any registers post_start_cmds.
        Should only be called once when the cluster is operational.
        """
        for cmd in self.post_start_cmds:
            try:
                self.dbg(f"Running post-start command: {cmd}")
                output = subprocess.check_output(cmd,
                                                 env=dict(os.environ,
                                                          **self.env),
                                                 shell=True)
                self.dbg('%s returned: %s' % (cmd, output))
            except subprocess.CalledProcessError as e:
                self.log('Failed to run %s' % (cmd))
                raise e

        # Avoid re-run
        self.post_start_cmds = list()

    def start(self):
        if self.state == 'started':
            raise Exception('%s already started' % self.name)

        if self.run():
            self.state = 'started'
            self.t_started = time.time()

    def pid(self):
        if self.proc is None:
            return 0
        return self.proc.pid

    def wait_stopped(self, timeout=30, force=False):
        """
        Wait for process to terminate.
        Application .state will be updated.

        @param force bool: Force process termination after @param timeout
        @param timeout float
        @returns True on succesful termination, else False.
        """
        t_end = time.time() + timeout
        while time.time() < t_end and (self.proc.poll() is None or
                                       self.proc.returncode is None):
            time.sleep(0.5)

        r = self.proc.wait()
        self.dbg("wait {} returned {}, wait_stopped returncode {}".format(
            self.proc.pid, r, self.proc.returncode))

        if self.proc.returncode is None:
            self.dbg('still alive')
            if force:
                self.dbg('forcing termination')
                os.killpg(os.getpgid(self.proc.pid), signal.SIGKILL)
                self.proc.wait()
            else:
                self.log('process did not terminate in %ds' % timeout)
                self.state = 'stale'
                return False

        self.state = 'stopped'
        self.t_stopped = time.time()
        return True

    def stop(self, wait_term=True, force=False):
        if self.state != 'started':
            return

        self.dbg('Stopping (pid {})'.format(self.proc.pid))
        try:
            os.killpg(os.getpgid(self.proc.pid), signal.SIGTERM)
        except OSError as e:
            self.log('killpg() failed: already dead? (%s): ignoring' % str(e))
            wait_term = False

        if wait_term:
            # Wait for termination
            self.wait_stopped(timeout=10, force=force)
        else:
            self.state = 'stopped'

        self.dbg('now %s, runtime %ds' % (self.state, self.runtime()))

        self.stdout_fd.close()
        self.stderr_fd.close()
        self.proc = None

    def status(self):
        if self.state == 'started' and self.proc is not None and self.proc.poll() is not None:  # noqa: E501
            r = self.proc.wait()
            if r != 0:
                self.log('process terminated: returncode %s' % (str(r)))
            else:
                self.dbg('process terminated: returncode %s' % (str(r)))
            self.state = 'stopped'
        return self.state

    def operational(self):
        return True  # Positive dummy: should be implemented by subclass

    def wait_operational(self, timeout=30):
        """ Wait for application to go operational """
        t_end = time.time() + timeout
        while time.time() < t_end:
            if self.status() == 'stopped':
                return False
            if self.operational():
                self.run_post_cmds()
                return True
            time.sleep(1.0)
        return False

    def deploy(self):
        """ Deploy application on node. NOT IMPLEMENTED """
        raise NotImplementedError('Deploy not implemented')

    def cleanup(self, keeptypes=['perm', 'log']):
        """ Remove all dirs and files created by App """
        if not self.do_cleanup:
            return
        self.dbg('Cleaning up %d path(s) (keeptypes=%s)' %
                 (len(self.paths), ','.join(keeptypes)))
        for p in self.paths:
            path = p['path']
            if not os.path.exists(path):
                continue
            self.dbg('Cleanup: %s (type %s, %s)' % (path, p['type'], p))
            if keeptypes is not None and p['type'] in keeptypes:
                continue
            try:
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)
            except Exception as e:
                self.dbg('Remove %s failed: %s: ignoring' % (path, str(e)))

    def runtime(self):
        if self.t_stopped < 1:
            return time.time() - self.t_started
        else:
            return self.t_stopped - self.t_started

    def __str__(self):
        return '{%s@%s:%s(%s)}' % (self.name, self.node.name,
                                   self.appid, self.state)


if __name__ == '__main__':
    pass



#!/usr/bin/env python
#

# Copyright (c) 2016-2019, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from trivup import trivup
from trivup.apps.KafkaBrokerApp import KafkaBrokerApp

import uuid
import requests
import subprocess


class SchemaRegistryApp (trivup.App):
    """ Confluent Schema Registry app.
        Depends on KafkaBrokerApp.
        Requires docker. """

    default_image = 'confluentinc/cp-schema-registry'
    default_version = 'latest'

    def __init__(self, cluster, conf=None, on=None):
        """
        @param cluster     Current cluster
        @param conf        Configuration dict, see below.
        @param on          Node name to run on

        Supported conf keys:
           * version - Confluent Platform version to use.
           * port_base - Low TCP port base to start allocating from (random)
           * image - docker image to use
           * conf - schema-registry docker image config strings (NOT USED)
        """
        super(SchemaRegistryApp, self).__init__(cluster, conf=conf, on=on)

        if self.conf.get('image', '') == '':
            self.conf['image'] = '{}:{}'.format(
                self.default_image,
                self.conf.get('version', self.default_version))

        self.conf['container_id'] = 'trivup_sr_%s' % str(uuid.uuid4())[0:7]
        kafka = cluster.find_app(KafkaBrokerApp)
        if kafka is None:
            raise Exception('KafkaBrokerApp required')

        bootstrap_servers = kafka.conf.get('docker_advertised_listeners')

        if bootstrap_servers is None:
            raise Exception('KafkaBrokerApp required')

        # Create listener
        port = trivup.TcpPortAllocator(self.cluster).next(
            self, self.conf.get('port_base', None))

        docker_args = ''
        if cluster.platform == 'linux':
            # Let container bind to host localhost
            self.conf['extport'] = port
            self.conf['intport'] = port
            docker_args = '--network=host'

        elif cluster.platform == 'darwin':
            # On OSX localhost binds are not possible, so set up a
            # port forwarding.
            self.conf['extport'] = port
            self.conf['intport'] = 8081
            docker_args = '-p %d:%d' % (self.conf['extport'],
                                        self.conf['intport'])

        # This is the listener address inside the docker container
        self.conf['listeners'] = 'http://0.0.0.0:%d' % self.conf.get('intport')
        # This is the listener address outside the docker container,
        # using port-forwarding
        self.conf['url'] = 'http://localhost:%d' % self.conf['extport']

        # Run in foreground.
        self.conf['start_cmd'] = 'docker run %s --name %s -e SCHEMA_REGISTRY_KAFKASTORE_BOOTSTRAP_SERVERS=%s  -e SCHEMA_REGISTRY_HOST_NAME=localhost   -e SCHEMA_REGISTRY_LISTENERS=%s  -e SCHEMA_REGISTRY_DEBUG=true %s' % (  # noqa: E501
            docker_args,
            self.conf.get('container_id'),
            bootstrap_servers,
            self.conf.get('listeners'),
            self.conf.get('image'))

        # Stop through docker
        self.conf['stop_cmd'] = 'docker stop %s' % \
                                self.conf.get('container_id')

    def operational(self):
        self.dbg('Checking if %s is operational' % self.get('url'))
        try:
            r = requests.head(self.get('url'), timeout=1.0)
            if r.status_code >= 200 and r.status_code < 300:
                return True
            raise Exception('status_code %d' % r.status_code)
        except Exception as e:
            self.dbg('%s check failed: %s' % (self.get('url'), e))
            return False

    def deploy(self):
        image = self.conf.get('image')
        self.dbg('Pulling docker image: %s' % image)
        subprocess.check_call('(docker images -q "%s" 2>/dev/null | grep -q ^.) || docker pull %s' % (image, image), shell=True)  # noqa: E501
        pass



#!/usr/bin/env python
#

# Copyright (c) 2021, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from trivup import trivup
from http.server import BaseHTTPRequestHandler, HTTPServer

from jwcrypto import jwk
import python_jwt as jwt
from datetime import datetime, timedelta
from threading import Lock

import json
import argparse
import requests

VALID_SCOPES = ['test', 'test-scope', 'api://1234-abcd/.default']

class WebServerHandler(BaseHTTPRequestHandler):
    def __init__(self):
        self._key = None
        self._public_keys = []
        self._mutex = Lock()

    def __call__(self, *args, **kwargs):
        """
        As it turns out, http.server.BaseHTTPRequestHandler
        ultimately inherits from socketserver.BaseRequestHandler,
        and socketserver.BaseRequestHandler.__init__() calls do_GET()
        and do_POST(). So if we add the super().__init__(*args, **kwargs)
        to the constructor of WebServerHandler, it means all the
        instance variables are initialized for each request.
        But the relevant parts of the JWT are cached in the broker.
        The only time the broker communicates with the OAuth/OIDC
        token provider is at a) startup, b) during a scheduled refresh,
        and c) when it finds a key ID that isn't cached.
        So we want to keep the same public key for each user in case
        the broker won't get the new public key immediately.
        So this method keeps the superclass "constructor"
        call out of "constructor" of this class which eliminates the
        possibility of dispatching a request (from the superclass's
        constructor) before the subclass's constructor is finished
        to avoid initializing the instance variables.
        Refer to https://stackoverflow.com/a/58909293
        """
        super().__init__(*args, **kwargs)

    def update_keys(self):
        self._mutex.acquire()
        if len(self._public_keys) == 0:
            public_key, key = self.generate_public_key()
            self._public_keys.append(json.loads(public_key))
            self._key = key
        self._mutex.release()

    def do_GET(self):
        if not self.path.endswith("/keys"):
            self.send_response(200)
            self.send_header('Content-Type', 'application/json')
            self.end_headers()
            message = "HTTP server for OAuth\n"
            message += "Example for token retrieval:\n"
            message += 'curl \
            -X POST \
            --url localhost:PORT/retrieve \
            -H "Accept: application/json" \
            -H "Content-Type: application/x-www-form-urlencoded" \
            -H "Authorization: Basic LW4gYWJjMTIzOlMzY3IzdCEK \
                (base64 string generated from CLIENT_ID:CLIENT_SECRET)" \
            -d "grant_type=client_credentials,scope=test-scope"'
            self.wfile.write(message.encode())
            return

        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.update_keys()
        keys = {"keys": self._public_keys}
        self.wfile.write(json.dumps(keys, indent=4).encode())

    def generate_token(self, lifetime, valid=True):
        self.update_keys()
        private_pem = self._key.export_to_pem(private_key=True, password=None)
        private_key = jwk.JWK.from_pem(private_pem)

        payload = {
            'exp': datetime.utcnow() + timedelta(days=0, seconds=lifetime),
            'iat': datetime.utcnow(),
            'iss': "issuer",
            'sub': "subject",
            'aud': 'api://default'
        }
        header = {
            "kid": "abcdefg"
        }

        token = jwt.generate_jwt(payload, private_key, 'RS256',
                                 timedelta(seconds=lifetime),
                                 other_headers=header)
        if not valid:
            token += "invalid"

        token_map = {"access_token": "%s" % token}
        return token_map

    def generate_public_key(self):

        key = jwk.JWK.generate(kty='RSA', size=2048, alg='RS256',
                               use='sig', kid="abcdefg")

        public_key = key.export_public()
        return (public_key, key)

    def valid_post_data(self, post_data):
        if post_data is None:
            self.send_error(400,
                            'grant_type=client_credentials and scope \
                             fields are required in data')
            return False

        post_data = post_data.decode("utf-8")

        if post_data == "grant_type=client_credentials":
            return True

        if not post_data.startswith("grant_type=client_credentials&scope="):
            self.send_error(400,
                            'format of data should be grant_type='
                            'client_credentials&scope=scope_value')
            return False

        index_begin_scope = len("grant_type=client_credentials&scope=")
        scope = post_data[index_begin_scope:]
        if scope not in VALID_SCOPES:
            self.send_error(400,
                            'Invalid scope, scope should be one of %s' %
                            VALID_SCOPES)
            return False

        return True

    def generate_valid_token_for_client(self):
        """
        Example usage:
        curl \
        -X POST \
        --url localhost:PORT/retrieve \
        -H "Accept: application/json" \
        -H "Content-Type: application/x-www-form-urlencoded" \
        -H "Authorization: Basic LW4gYWJjMTIzOlMzY3IzdCEK \
            (base64 string generated from CLIENT_ID:CLIENT_SECRET)"
        -d "grant_type=client_credentials,scope=test-scope"
        """
        if self.headers.get('Content-Length', None) is None:
            self.send_error(400, 'Content-Length field is required')
            return

        if self.headers.get('Content-Type', None) is None:
            self.send_error(400, 'Content-Type field is required')
            return

        content_length = int(self.headers['Content-Length'])
        post_data = self.rfile.read(content_length)

        if self.headers.get('Authorization', None) is None:
            self.send_error(400, 'Authorization field is required')
            return

        if self.headers.get('Accept', None) != "application/json":
            self.send_error(400, 'Accept field should be "application/json"')
            return

        if not self.valid_post_data(post_data):
            return

        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        token = self.generate_token(4)
        self.wfile.write(json.dumps(token, indent=4).encode())

    def generate_badformat_token_for_client(self):
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()

        token = self.generate_token(30, False)
        self.wfile.write(json.dumps(token, indent=4).encode())

    def generate_expired_token_for_client(self):
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()

        token = self.generate_token(-1)
        self.wfile.write(json.dumps(token, indent=4).encode())

    def do_POST(self):
        if self.path.endswith("/retrieve"):
            self.generate_valid_token_for_client()
        elif self.path.endswith("/retrieve/badformat"):
            self.generate_badformat_token_for_client()
        elif self.path.endswith("/retrieve/expire"):
            self.generate_expired_token_for_client()
        else:
            self.send_error(404, 'URL is not valid: %s' % self.path)


class OauthbearerOIDCHttpServer():
    def run_http_server(self, port):
        handler = WebServerHandler()
        server = HTTPServer(('localhost', port), handler)
        server.serve_forever()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Trivup Oauthbearer OIDC \
                                                  HTTP server')
    parser.add_argument('--port', type=int, dest='port',
                        required=True,
                        help='Port at which OauthbearerOIDCApp \
                              should be bound')
    args = parser.parse_args()

    http_server = OauthbearerOIDCHttpServer()
    http_server.run_http_server(args.port)


class OauthbearerOIDCApp (trivup.App):
    """ Oauth/OIDC app, run a http server """
    def __init__(self, cluster, conf=None, on=None):
        """
        @param cluster     Current cluster.
        @param conf        Configuration dict.
               port        Port at which OauthbearerOIDCApp should be bound
                           (optional). A (random) free port will be chosen
                           otherwise.
        @param on          Node name to run on.
        """
        super(OauthbearerOIDCApp, self).__init__(cluster, conf=conf, on=on)
        self.conf['port'] = trivup.TcpPortAllocator(self.cluster).next(
            self, port_base=self.conf.get('port', None))
        self.conf['valid_url'] = 'http://localhost:%d/retrieve' % \
            self.conf['port']
        self.conf['badformat_url'] = 'http://localhost:%d/retrieve/badformat' \
            % self.conf['port']
        self.conf['expired_url'] = 'http://localhost:%d/retrieve/expire' % \
            self.conf['port']
        self.conf['jwks_url'] = 'http://localhost:%d/keys' % self.conf['port']
        self.conf['sasl_oauthbearer_method'] = 'OIDC'
        self.conf['sasl_oauthbearer_client_id'] = '123'
        self.conf['sasl_oauthbearer_client_secret'] = 'abc'
        self.conf['sasl_oauthbearer_scope'] = 'test'
        self.conf['sasl_oauthbearer_extensions'] = \
            'ExtensionworkloadIdentity=develC348S,Extensioncluster=lkc123'

    def start_cmd(self):
        return "python3 -m trivup.apps.OauthbearerOIDCApp --port %d" \
               % self.conf['port']

    def operational(self):
        self.dbg('Checking if %s is operational' % self.get('valid_url'))
        try:
            r = requests.get(self.get('valid_url'))
            if r.status_code == 200:
                return True
            raise Exception('status_code %d' % r.status_code)
        except Exception as e:
            self.dbg('%s check failed: %s' % (self.get('valid_url'), e))
            return False

    def deploy(self):
        pass



#!/usr/bin/env python
#

# Copyright (c) 2016-2022, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# Provides a Kafka cluster with the following components:
#  * ZooKeeperApp (1)
#  * KafkaBrokerApp brokers (broker_cnt=3)
#  * SslApp (optional, if with_ssl=True)
#  * KerberosKdcApp (optional, sasl.mechanism=GSSAPI,
#                    cross-realm if realm_cnt=2)
#  * SchemaRegistryApp (optional, if with_sr=True)
#  * OauthbearerOIDCApp (optional, if oidc=True)
#
# cluster.env (dict) will contain:
#      TRIVUP_ROOT
#      ZK_ADDRESS  (unless --kraft)
#      BROKERS
#      BROKER_PID_<nodeid>
#      KAFKA_PATH  (path to kafka package root directory)
#      SR_URL      (if with_sr)
#      SSL_(ca|pub|priv)_.. (if with_ssl, paths to certs and keys)
#      SSL_password         (if with_ssl, key password)
#      KRB5CCNAME, KRB5_COFNIG, KRB5_KDC_PROFILE (if GSSAPI enabled)
#
# See conf dict structure below.

from trivup.trivup import Cluster, Allocator, TcpPortAllocator
from trivup.apps.ZookeeperApp import ZookeeperApp
from trivup.apps.KafkaBrokerApp import KafkaBrokerApp
from trivup.apps.KerberosKdcApp import KerberosKdcApp
from trivup.apps.SchemaRegistryApp import SchemaRegistryApp
from trivup.apps.SslApp import SslApp
from trivup.apps.OauthbearerOIDCApp import OauthbearerOIDCApp

from copy import deepcopy

import os
import argparse
import subprocess
import copy


class KafkaCluster(object):
    # conf dict structure with defaults:
    # commented-out fields are not defaults but show what is available.
    default_conf = {
        'version': '2.8.0',     # Apache Kafka version
        'cp_version': '6.1.0',  # Confluent Platform version (for SR)
        'broker_cnt': 3,
        'sasl_mechanism': '',   # GSSAPI, PLAIN, SCRAM-.., ...
        'realm_cnt': 1,
        'krb_renew_lifetime': 30,
        'krb_ticket_lifetime': 120,
        # KRaft (No zookeeper). Requires AK >=2.8
        'kraft': False,
        # SASL PLAIN/SCRAM
        'sasl_users': 'testuser=testpass',
        # With SSL
        'with_ssl': False,
        # With SchemaRegistry
        'with_sr': False,
        # Debug trivup
        'debug': False,
        # Cleanup
        'cleanup': True,
        'oidc': False,
        # Additional broker server.properties configuration
        # 'broker_conf': ['connections.max.idle.ms=1234', ..]
    }

    def __init__(self, **kwargs):
        """ Create and start a KafkaCluster.
            See default_conf above for parameters. """
        super(KafkaCluster, self).__init__()

        conf = kwargs
        self.conf = deepcopy(self.default_conf)
        if conf is not None:
            self.conf.update(conf)

        self.version = self.conf.get('version')
        self.version_num = [int(x) for x in self.version.split('.')][:3]
        self.kraft = self.conf.get('kraft')

        # Create trivup Cluster
        self.cluster = Cluster(
            self.__class__.__name__,
            os.environ.get('TRIVUP_ROOT', 'tmp-%s' % self.__class__.__name__),
            debug=bool(self.conf.get('debug', False)),
            cleanup=bool(self.conf.get('cleanup', True)))

        self._client_conf = dict()
        self.env = dict()

        self.sasl_mechanism = self.conf.get('sasl_mechanism')

        # Add OIDC server app
        if bool(self.conf.get('oidc', False)):
            self.oidc = OauthbearerOIDCApp(self.cluster)
            if not self.sasl_mechanism:
                self.sasl_mechanism = 'OAUTHBEARER'
            elif self.sasl_mechanism.upper() != 'OAUTHBEARER':
                raise RuntimeError(f"OIDC requires sasl.mechanism OAUTHBEARER, not '{self.sasl_mechanism}'")

        # Generate SSL certs if enabled
        if bool(self.conf.get('with_ssl')):
            self.ssl = SslApp(self.cluster, self.conf)
        else:
            self.ssl = None

        # Map mechanism and SSL to security protocol
        self.security_protocol = {(True, True): 'SASL_SSL',
                                  (True, False): 'SASL_PLAINTEXT',
                                  (False, True): 'SSL',
                                  (False, False): 'PLAINTEXT'}[
                                      (bool(self.sasl_mechanism),
                                       bool(self.ssl is not None))]

        if not self.kraft:
            # Create single ZK for the cluster (don't start yet)
            self.zk = ZookeeperApp(self.cluster)
        else:
            self.zk = None
            # Allocate (but don't use) a dummy appid so that the brokers get
            # the same appid/nodeid for both KRaft and ZK modes.
            Allocator(self.cluster).next()

        # Broker configuration
        broker_cnt = int(self.conf.get('broker_cnt'))
        self.broker_conf = {'replication_factor': min(3, broker_cnt),
                            'num_partitions': 4,
                            'version': self.version,
                            'sasl_mechanisms': self.sasl_mechanism,
                            'sasl_users': self.conf.get('sasl_users'),
                            'conf': self.conf.get('broker_conf', []),
                            'kafka_path': self.conf.get('kafka_path', None)}

        # Start Kerberos KDCs if GSSAPI (Kerberos) is configured
        if self.sasl_mechanism == 'GSSAPI':
            self._setup_kerberos()
            self.broker_conf['realm'] = self.broker_realm

        self.broker_conf['listener_host'] = 'localhost'
        # Create brokers (don't start yet)
        self.brokers = dict()
        for n in range(0, broker_cnt):
            bconf = copy.deepcopy(self.broker_conf)
            if self.version_num >= [2, 4, 0]:
                # Configure rack & replica selector if broker supports
                # fetch-from-follower
                bconf.update(
                    { 'conf': [
                        'broker.rack=RACK${appid}',
                        'replica.selector.class=org.apache.kafka.common.replica.RackAwareReplicaSelector']})  # noqa: E501

            broker = KafkaBrokerApp(self.cluster, bconf)
            self.brokers[broker.appid] = broker

        # Generate bootstrap servers list
        all_listeners = (','.join(self.cluster.get_all(
            'advertised_listeners', '', KafkaBrokerApp))).split(',')
        self.bootstrap_servers = ','.join(
            [x for x in all_listeners if x.startswith(self.security_protocol)])

        assert len(self.bootstrap_servers) >= broker_cnt, \
            "{} < {} expected bootstrap servers".format(
                len(self.bootstrap_servers), broker_cnt)

        # Create SchemaRegistry if enabled
        if bool(self.conf.get('with_sr', False)):
            self.sr = SchemaRegistryApp(
                self.cluster, {'version': self.conf.get('cp_version')})
            self.env['SR_URL'] = self.sr.get('url')

        # Create librdkafka client configuration
        self._setup_client_conf()

        # Deploy cluster
        self.cluster.deploy()

        # Start cluster
        self.start()

    def __del__(self):
        """ Destructor: forcibly stop the cluster """
        self.stop(force=True)

    def _setup_env(self):
        """ Set up convenience envs """
        self.env['KAFKA_PATH'] = self.cluster.find_app(KafkaBrokerApp).get('destdir')  # noqa: E501
        if not self.kraft:
            self.env['ZK_ADDRESS'] = self.zk.get('address')
        self.env['BROKERS'] = self.bootstrap_servers
        self.env['KAFKA_VERSION'] = self.version
        self.env['TRIVUP_ROOT'] = self.cluster.instance_path()

        # Add each broker pid as an env so they can be killed indivdidually.
        for b in self.cluster.find_apps(KafkaBrokerApp, 'started'):
            self.env['BROKER_PID_%d' % b.appid] = str(b.proc.pid)

    def _setup_client_conf(self):
        """ Set up librdkafka client configuration """
        self._client_conf['bootstrap.servers'] = self.bootstrap_servers
        self._client_conf['broker.address.family'] = 'v4'
        if self.security_protocol != 'PLAINTEXT':
            self._client_conf['security.protocol'] = self.security_protocol

        broker_version = self.conf.get('version')
        brver = broker_version.split('.')
        if brver[0] == 0 and brver[1] < 10:
            self._client_conf['broker.version.fallback'] = broker_version
            self._client_conf['api.version.request'] = 'false'

        # Client SASL configuration
        if self.sasl_mechanism:
            self._client_conf['sasl.mechanism'] = self.sasl_mechanism

            if self.sasl_mechanism == 'PLAIN' or \
               self.sasl_mechanism.find('SCRAM') != -1:
                # Use first user as SASL user/pass
                for up in self.conf.get('sasl_users', '').split(','):
                    u, p = up.split('=')
                    self._client_conf['sasl.username'] = u
                    self._client_conf['sasl.password'] = p
                    break

            elif self.sasl_mechanism == 'OAUTHBEARER':
                if self.oidc is not None:
                    self._client_conf['sasl.oauthbearer.method'] = 'OIDC'
                    self._client_conf['sasl.oauthbearer.token.endpoint.url'] = self.oidc.get('valid_url')  # noqa: E501
                    self._client_conf['sasl.oauthbearer.client.id'] = '123'
                    self._client_conf['sasl.oauthbearer.client.secret'] = 'abc'
                    self._client_conf['sasl.oauthbearer.scope'] = 'test'
                    self._client_conf['sasl.oauthbearer.extensions'] = \
                        'ExtensionworkloadIdentity=develC348S,Extensioncluster=lkc123'
                else:
                    self._client_conf['enable.sasl.oauthbearer.unsecure.jwt'] = True  # noqa: E501
                    self._client_conf['sasl.oauthbearer.config'] = \
                        'scope=requiredScope principal=admin'

        # Client SSL configuration
        if self.ssl is not None:
            key = self.ssl.create_cert('client')
            self._client_conf['ssl.ca.location'] = self.ssl.ca['pem']
            self._client_conf['ssl.certificate.location'] = key['pub']['pem']
            self._client_conf['ssl.key.location'] = key['priv']['pem']
            self._client_conf['ssl.key.password'] = key['password']

            # Add envs pointing out locations of the generated certs
            for k, v in self.ssl.ca.items():
                self.env['SSL_ca_{}'.format(k)] = v

            for k, v in self.ssl.unused_ca.items():
                self.env['SSL_unused_ca_{}'.format(k)] = v

            self.env['SSL_all_cas_pem'] = self.ssl.all_cas['pem']

            # Set envs for all generated keys so tests can find them.
            for k, v in key.items():
                if type(v) is dict:
                    for k2, v2 in v.items():
                        # E.g. "SSL_priv_der=path/to/librdkafka-priv.der"
                        self.env['SSL_{}_{}'.format(k, k2)] = v2
                else:
                    self.env['SSL_{}'.format(k)] = v

    def _setup_kerberos(self):
        """ Set up Kerberos KDCs """

        # Create KDCs for each realm.
        # First realm will be the default / broker realm.
        #
        realm_cnt = int(self.conf.get('realm_cnt', 1))
        # No point in having more than two realms
        assert realm_cnt > 0 and realm_cnt < 3
        realms = ['REALM{}.TRIVUP'.format(x + 1) for x in range(0, realm_cnt)]

        # Pre-Allocate ports for the KDCs so they can reference eachother
        # in the krb5.conf configuration.
        kdc_ports = {x: TcpPortAllocator(
            self.cluster).next("dummy") for x in realms}

        # Set up realm=kdc:port cross-realm mappings
        cross_realms = ",".join(["{}={}:{}".format(
            x, self.cluster.get_node().name, kdc_ports[x]) for x in realms])

        kdcs = dict()
        for realm in realms:
            kdc = KerberosKdcApp(
                self.cluster, realm,
                conf={'port': kdc_ports[realm],
                      'cross_realms': cross_realms,
                      'renew_lifetime':
                      int(self.conf.get('krb_renew_lifetime')),
                      'ticket_lifetime':
                      int(self.conf.get('krb_ticket_lifetime'))})
            # Kerberos needs to be started prior to Kafka so that principals
            # and keytabs are available at the time of Kafka config generation.
            kdc.start()
            kdcs[realm] = kdc

        self.broker_realm = realms[0]
        self.client_realm = realms[-1]
        self.broker_kdc = kdcs[self.broker_realm]
        self.client_kdc = kdcs[self.client_realm]

        # Add cross-realm TGTs
        if realm_cnt > 1:
            KerberosKdcApp.add_cross_realm_tgts(kdcs)

        # Add client envs and configuration
        self.env['KRB5CCNAME'] = self.client_kdc.mkpath('krb5cc')
        self.env['KRB5_CONFIG'] = self.client_kdc.conf['krb5_conf']
        self.env['KRB5_KDC_PROFILE'] = self.client_kdc.conf['kdc_conf']
        principal, keytab = self.client_kdc.add_principal('admin')

        self._client_conf['sasl.kerberos.keytab'] = keytab
        self._client_conf['sasl.kerberos.principal'] = principal.split('@')[0]
        # Refresh ticket 60s before renew timeout.
        self._client_conf['sasl.kerberos.min.time.before.relogin'] = \
            max(1, int(self.conf.get('krb_renew_lifetime')) - 60) * 1000

    def start(self, timeout=0):
        """ Start cluster """
        self.cluster.start()

        # Set up additional convenience envs
        self._setup_env()

        if timeout > 0:
            self.wait_operational(timeout)

    def stop(self, cleanup=True, keeptypes=['log'], force=False, timeout=0):
        """ Stop cluster and clean up """
        self.cluster.stop(force=True)
        if timeout > 0:
            self.cluster.wait_stopped(timeout)
        if cleanup:
            self.cluster.cleanup(keeptypes)

    def stopped(self):
        """ Returns True when all components of the cluster are stopped """
        return len([x for x in self.cluster.apps if
                    x.status() == 'stopped']) == len(self.cluster.apps)

    def wait_operational(self, timeout=60):
        """ Wait for cluster to go operational """
        if not self.cluster.wait_operational(timeout):
            self.stop(force=True)
            raise Exception(
                "Cluster {} did not go operational, see logs in {}/{}".format(
                    self.cluster.name, self.cluster.root_path,
                    self.cluster.instance))

    def stop_broker(self, broker_id):
        """ Stop single broker """
        broker = self.brokers.get(broker_id, None)
        if broker is None:
            raise LookupError("Unknown broker id {}".format(broker_id))
        broker.stop()

    def start_broker(self, broker_id):
        """ Start single broker """
        broker = self.brokers.get(broker_id, None)
        if broker is None:
            raise LookupError("Unknown broker id {}".format(broker_id))
        broker.start()

    def interactive(self, cmd=None):
        """ Execute an interactive shell that has all the
            environment variables set. """

        if cmd is None:
            print('# Interactive mode')

        print("# - Waiting for cluster to go operational in {}/{}".format(
            self.cluster.root_path, self.cluster.instance))
        kc.wait_operational()

        env = self.env.copy()

        # Avoids 'dumb' terminal mode
        env['TERM'] = os.environ.get('TERM', 'vt100')

        # Write librdkafka client configuration to file.
        cf_path = self.cluster.mkpath('rdkafka.conf', in_instance=True)
        self.write_client_conf(cf_path)
        env['RDKAFKA_TEST_CONF'] = cf_path
        print("# - Client configuration in {}".format(cf_path))
        print("# - Connect to cluster with bootstrap.servers {}".format(
            self.bootstrap_servers))

        # Prefix the standard prompt with cluster info.
        if cmd is None:
            pfx = '[TRIVUP:{}@{}] '.format(self.cluster.name, self.version)
            fullcmd = 'bash --rcfile <(cat ~/.bashrc; echo \'PS1="{}$PS1"\') -i'.format(pfx)  # noqa: E501
            print("# - You're now in an interactive sub-shell, type 'exit' "
                  "to stop the cluster and exit back to your shell.\n")
            retcode = subprocess.call(fullcmd, env=env, shell=True,
                                      executable='/bin/bash')

        else:
            fullcmd = cmd
            print("# - Executing: {}".format(fullcmd))
            retcode = subprocess.call(fullcmd, env=env, shell=True)

        if retcode != 0:
            print("# - Shell exited with returncode {}: {}".format(
                retcode, fullcmd))

    def client_conf(self):
        """ Get a dict copy of the client configuration """
        return deepcopy(self._client_conf)

    def write_client_conf(self, path, additional_blob=None):
        """ Write client configuration (librdkafka) to @param path """
        with open(path, "w") as f:
            for k, v in self._client_conf.items():
                f.write(str('%s=%s\n' % (k, v)))
            if additional_blob is not None:
                f.write(str('#\n# Additional configuration:'))
                f.write(str(additional_blob))


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Trivup KafkaCluster')
    parser.add_argument('--debug', action='store_true', dest='debug',
                        default=False, help='Enable trivup debugging')
    parser.add_argument('--kraft', dest='kraft', action='store_true',
                        default=KafkaCluster.default_conf['kraft'],
                        help='KRaft mode (no Zookeeper). Requires >= AK 2.8')
    parser.add_argument('--sasl', dest='sasl', type=str,
                        default=KafkaCluster.default_conf['sasl_mechanism'],
                        help='SASL mechanism (PLAIN, SCRAM-SHA-nnn, GSSAPI, ' +
                        'OAUTHBEARER)')
    parser.add_argument('--ssl', dest='ssl', action='store_true',
                        default=KafkaCluster.default_conf['with_ssl'],
                        help='Enable SSL')
    parser.add_argument('--sr', dest='sr', action='store_true',
                        default=KafkaCluster.default_conf['with_sr'],
                        help='Enable SchemaRegistry')
    parser.add_argument('--brokers', dest='broker_cnt', type=int,
                        default=KafkaCluster.default_conf['broker_cnt'],
                        help='Number of Kafka brokers')
    parser.add_argument('--version', dest='version', type=str,
                        default=KafkaCluster.default_conf['version'],
                        help='Apache Kafka version')
    parser.add_argument('--cpversion', dest='cp_version', type=str,
                        default=KafkaCluster.default_conf['cp_version'],
                        help='Confluent Platform version (for Schema-Registry)')  # noqa: E501
    parser.add_argument('--no-cleanup', dest='no_cleanup', action='store_true',
                        default=False,
                        help='Don\'t clean up logs, etc, after cluster is destroyed')
    parser.add_argument('--cmd', type=str, dest='cmd', default=None,
                        help='Command to execute instead of interactive shell')
    parser.add_argument('--kafka-src', dest='kafka_src', type=str,
                        default=None,
                        help='Existing Kafka source directory to use ' +
                        'instead of downloaded release. ' +
                        'Requires --version trunk.')
    parser.add_argument('--oidc', dest='oidc',
                        action='store_true',
                        default=KafkaCluster.default_conf['oidc'],
                        help='Enable Oauthbearer OIDC JWT server')

    args = parser.parse_args()

    conf = {'debug': args.debug,
            'kraft': args.kraft,
            'version': args.version,
            'cp_version': args.cp_version,
            'sasl_mechanism': args.sasl,
            'with_ssl': args.ssl,
            'with_sr': args.sr,
            'broker_cnt': args.broker_cnt,
            'kafka_path': args.kafka_src,
            'cleanup': not args.no_cleanup,
            'oidc': args.oidc}

    kc = KafkaCluster(**conf)

    kc.interactive(args.cmd)

    print("# Stopping cluster in {}/{}".format(kc.cluster.root_path,
                                               kc.cluster.instance))

    kc.stop()



#!/usr/bin/env python
#

# Copyright (c) 2016-2021, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from trivup import trivup
from trivup.apps.KerberosKdcApp import KerberosKdcApp
from trivup.apps.SslApp import SslApp
from trivup.apps.OauthbearerOIDCApp import OauthbearerOIDCApp

from string import Template
import contextlib
import os
import socket
import time


class KafkaBrokerApp (trivup.App):
    """ Kafka broker app
        Depends on ZookeeperApp (unless KRaft mode) """
    def __init__(self, cluster, conf=None, on=None):
        """
        @param cluster     Current cluster
        @param conf        Configuration dict, see below.
        @param on          Node name to run on

        Supported conf keys:
           * version - Kafka version to use, will build 'trunk' from
                       kafka_path, otherwise the version is taken to be a
                       formal release which will be downloaded and deployed.
           * listeners - CSV list of listener types:
                         PLAINTEXT,SSL,SASL,SASL_SSL
           * listener_host - alternative listener host instead of
                             node name (e.g., '*')
           * advertised_hostname - hostname to use for advertised.listeners
                                   (defaults to 'on' node)
           * sasl_mechanisms - CSV list of SASL mechanisms to enable:
                               GSSAPI,PLAIN,SCRAM-SHA-n,OAUTHBEARER
                               SASL listeners will be added automatically.
                               KerberosKdcApp is required for GSSAPI.
           * sasl_users - CSV list of SASL PLAIN/SCRAM of user=pass for
                          authenticating clients
           * ssl_client_auth - ssl.client.auth broker property (def: required)
           * auto_create_topics - Enable topic auto-creation (true)
           * num_partitions - Topic auto-create partition count (3)
           * replication_Factor - Topic auto-create replication factor (1)
           * port_base - Low TCP port base to start allocating from (random)
           * port - Alias for port_base
           * kafka_path - Path to Kafka build tree (for trunk usage)
           * fdlimit - RLIMIT_NOFILE (or "max") (default: max)
           * conf - arbitary server.properties config as a list of strings.
           * realm - Kerberos realm to use when sasl_mechanisms contains GSSAPI
        """
        super(KafkaBrokerApp, self).__init__(cluster, conf=conf, on=on)

        # Kafka repo uses SVN nomenclature
        if self.conf['version'] == 'master':
            self.conf['version'] = 'trunk'

        if self.conf['version'] == 'trunk':
            self.version = [9, 9, 9]
        else:
            self.version = [int(x) for x in
                            self.conf['version'].split('.')][:3]

        self.zk = cluster.find_app('ZookeeperApp')
        if self.zk is None:
            # If AK >=2.8 we can run in KRaft-mode without ZK
            if self.version < [2, 8, 0]:
                raise Exception('ZookeeperApp or AK >=2.8 required')
            self.kraft = True
        else:
            self.kraft = False

        if 'fdlimit' not in self.conf:
            self.conf['fdlimit'] = 'max'

        listener_host = self.conf.get('listener_host',
                                      self.conf.get('nodename'))
        self.conf['listener_host'] = listener_host

        # Kafka Configuration properties
        self.conf['log_dirs'] = self.create_dir('logs')
        if 'num_partitions' not in self.conf:
            self.conf['num_partitions'] = 3

        if self.zk is not None:
            self.conf['zk_connect'] = self.zk.get('address', None)

        if 'replication_factor' not in self.conf:
            self.conf['replication_factor'] = 1

        # Kafka paths
        if self.conf.get('kafka_path', None) is not None:
            self.conf['destdir'] = self.conf['kafka_path']
            self.conf['bindir'] = os.path.join(self.conf['destdir'], 'bin')
            start_sh = os.path.join(self.conf['bindir'],
                                    'kafka-server-start.sh')
            kafka_configs_sh = os.path.join(self.conf['bindir'],
                                            'kafka-configs.sh')
        else:
            start_sh = 'kafka-server-start.sh'
            kafka_configs_sh = 'kafka-configs.sh'

        # Arbitrary (non-template) configuration statements
        conf_blob = self.conf.get('conf', list())
        jaas_blob = list()

        #
        # Configure listener types, SSL and SASL, CONTROLLER, etc.
        #
        listener_types = self.conf.get('listeners', 'PLAINTEXT').split(',')

        if self.kraft:
            conf_blob.append('process.roles=broker,controller')
            conf_blob.append('controller.listener.names=CONTROLLER')
            listener_types.append('CONTROLLER')
        else:
            conf_blob.append('zookeeper.connect={}'.format(
                self.conf.get('zk_connect')))

        # SASL support
        sasl_mechs = [x for x in self.conf.get('sasl_mechanisms', '').
                      replace(' ', '').split(',') if len(x) > 0]
        if len(sasl_mechs) > 0:
            listener_types.append('SASL_PLAINTEXT')

        # SSL support
        ssl = cluster.find_app(SslApp)
        if ssl is not None:
            # Add SSL listener_types
            listener_types.append('SSL')
            if len(sasl_mechs) > 0:
                listener_types.append('SASL_SSL')

        listener_map = 'listener.security.protocol.map=' + \
            'PLAINTEXT:PLAINTEXT,SSL:SSL,SASL_PLAINTEXT:' + \
            'SASL_PLAINTEXT,SASL_SSL:SASL_SSL'

        can_docker = self.conf['version'] == 'trunk' or \
            int(self.conf['version'].split('.')[0]) > 0
        if can_docker:
            # Map DOCKER listener to PLAINTEXT security protocol
            listener_map += ',DOCKER:PLAINTEXT'

        if self.kraft:
            listener_map += ',CONTROLLER:PLAINTEXT'

        conf_blob.append(listener_map)

        def sort_listener(a):
            """ Sort listener_types list so that the PLAINTEXTs are first,
                since the first listener is used by operational(). """
            if a.startswith('PLAINTEXT'):
                return 0
            else:
                return 1

        # Allocate a port for each listener type
        ports = [(x, trivup.TcpPortAllocator(self.cluster).next(
            self, self.conf.get('port_base', self.conf.get('port', None))))
                 for x in sorted(set(listener_types), key=sort_listener)]
        self.conf['port'] = ports[0][1]  # "Default" port

        if can_docker:
            # Add docker listener to allow services (e.g, SchemaRegistry) in
            # docker-containers to reach the on-host Kafka.
            docker_port = trivup.TcpPortAllocator(self.cluster).next(self)
            docker_host = '%s:%d' % (cluster.get_docker_host(), docker_port)

        self.conf['address'] = '%s:%d' % (listener_host, self.conf['port'])
        # Create a listener for each port
        listeners = ['%s://%s:%d' % (x[0], "0.0.0.0", x[1]) for x in ports]
        if can_docker:
            listeners.append('%s://%s:%d' % ('DOCKER', "0.0.0.0", docker_port))
        self.conf['listeners'] = ','.join(listeners)
        if 'advertised_hostname' not in self.conf:
            self.conf['advertised_hostname'] = self.conf['nodename']
        advertised_listeners = ['%s://%s:%d' %
                                (x[0], self.conf['advertised_hostname'], x[1])
                                for x in ports if x[0] != 'CONTROLLER']
        if can_docker:
            # Expose service to docker containers as well.
            advertised_listeners.append('DOCKER://%s' % docker_host)
            self.conf['docker_advertised_listeners'] = 'PLAINTEXT://%s' % \
                docker_host
        self.conf['advertised.listeners'] = ','.join(advertised_listeners)
        self.conf['advertised_listeners'] = self.conf['advertised.listeners']
        if self.kraft:
            self.conf['controller_listener'] = \
                [x for x in listeners if x.startswith('CONTROLLER://')][0]

        self.conf['auto_create_topics'] = self.conf.get('auto_create_topics',
                                                        'true')
        self.dbg('Listeners: %s' % self.conf['listeners'])
        self.dbg('Advertised Listeners: %s' %
                 self.conf['advertised.listeners'])

        self._add_simple_authorizer(conf_blob)

        if len(sasl_mechs) > 0:
            self.dbg('SASL mechanisms: %s' % sasl_mechs)
            jaas_blob.append('KafkaServer {')

            conf_blob.append('sasl.enabled.mechanisms=%s' %
                             ','.join(sasl_mechs))
            # Handle PLAIN and SCRAM-.. the same way
            for mech in sasl_mechs:
                if mech.find('SCRAM') != -1:
                    plugin = 'scram.Scram'
                elif mech == 'PLAIN':
                    plugin = 'plain.Plain'
                else:
                    continue

                sasl_users = self.conf.get('sasl_users', '')
                if len(sasl_users) == 0:
                    self.log(('WARNING: No sasl_users configured for %s, '
                              'expected CSV of user=pass,..') % plugin)
                else:
                    jaas_blob.append(('org.apache.kafka.common.security.'
                                      '%sLoginModule required debug=true') %
                                     plugin)
                    for up in sasl_users.split(','):
                        u, p = up.split('=')
                        if plugin == 'plain.Plain':
                            jaas_blob.append('  user_%s="%s"' % (u, p))
                        elif plugin == 'scram.Scram':
                            jaas_blob.append('  username="%s" password="%s"' %
                                             (u, p))
                            # SCRAM users are set up using kafka-configs.sh
                            self.post_start_cmds.append((
                                'JMX_PORT="" %s --bootstrap-server %s '
                                '--alter --add-config '
                                '\'%s=[iterations=4096,password=%s]\' '
                                '--entity-type users --entity-name \'%s\'') % (
                                    kafka_configs_sh,
                                    self.conf['advertised_listeners'],
                                    mech, p, u))

                    jaas_blob[-1] += ';'

            if 'GSSAPI' in sasl_mechs:
                conf_blob.append('sasl.kerberos.service.name=%s' % 'kafka')
                realm = self.conf.get('realm', None)
                if realm is None:
                    kdc = self.cluster.find_app(KerberosKdcApp)
                    self.conf['realm'] = kdc.conf.get('realm', None)
                else:
                    kdc = self.cluster.find_app(KerberosKdcApp,
                                                ('realm', realm))
                    # If a realm was specified it is most likely because
                    # we're operating in a cross-realm scenario.
                    # Add a principal mapping client principals without
                    # hostname ("admin" rather than "admin/localhost")
                    # to a local user.
                    # This is not compatible with "admin/localhost" principals.
                    self._add_simple_authorizer(conf_blob)
                    conf_blob.append('allow.everyone.if.no.acl.found=true')
                    conf_blob.append('sasl.kerberos.principal.to.local.rules=RULE:[1:admin](.*)s/^.*/admin/')  # noqa: E501

                assert kdc is not None, \
                    "No KerberosKdcApp found (realm={})".format(realm)
                self.env_add('KRB5_CONFIG', kdc.conf['krb5_conf'])
                self.env_add('KAFKA_OPTS', '-Djava.security.krb5.conf=%s' %
                             kdc.conf['krb5_conf'])
                self.env_add('KAFKA_OPTS', '-Dsun.security.krb5.debug=true')
                self.kerberos_principal, self.kerberos_keytab = kdc.add_principal('kafka', self.conf['advertised_hostname'])  # noqa: E501
                jaas_blob.append('com.sun.security.auth.module.Krb5LoginModule required')  # noqa: E501
                jaas_blob.append('useKeyTab=true storeKey=true doNotPrompt=true')  # noqa: E501
                jaas_blob.append('keyTab="%s"' % self.kerberos_keytab)
                jaas_blob.append('debug=true')
                jaas_blob.append('principal="%s";' % self.kerberos_principal)

            if 'OAUTHBEARER' in sasl_mechs:
                oidcapp = self.cluster.find_app(OauthbearerOIDCApp)
                if oidcapp is not None:
                    assert self.version >= [3, 1, 0], "OIDC requires Apache Kafka 3.1 or later"
                    # Use the OIDC method.
                    conf_blob.append('listener.name.sasl_plaintext.oauthbearer.sasl.server.callback.handler.class=org.apache.kafka.common.security.oauthbearer.secured.OAuthBearerValidatorCallbackHandler')
                    conf_blob.append('listener.name.sasl_plaintext.oauthbearer.sasl.oauthbearer.jwks.endpoint.url=%s' % oidcapp.conf['jwks_url'])
                    conf_blob.append('listener.name.sasl_plaintext.oauthbearer.sasl.oauthbearer.scope.claim.name=scp')
                    conf_blob.append('listener.name.sasl_plaintext.oauthbearer.sasl.jaas.config=org.apache.kafka.common.security.oauthbearer.OAuthBearerLoginModule required unsecuredLoginStringClaim_sub="unused";')
                    conf_blob.append('listener.name.sasl_plaintext.oauthbearer.sasl.oauthbearer.expected.audience=api://default')
                    conf_blob.append('security.inter.broker.protocol=PLAINTEXT')
                    conf_blob.append('sasl.enabled.mechanisms=OAUTHBEARER')
                else:
                    # Use the unsecure JSON web token.
                    # Client should be configured with
                    # 'sasl.oauthbearer.config=scope=requiredScope principal=
                    # admin'
                    # Change requiredScope to something else to trigger auth
                    # error.
                    conf_blob.append('super.users=User:admin')
                    conf_blob.append('allow.everyone.if.no.acl.found=true')
                    self._add_simple_authorizer(conf_blob)
                    jaas_blob.append('org.apache.kafka.common.security.oauthbearer.OAuthBearerLoginModule required')  # noqa: E501
                    jaas_blob.append('  unsecuredLoginLifetimeSeconds="3600"')
                    jaas_blob.append('  unsecuredLoginStringClaim_sub="admin"')
                    jaas_blob.append('  unsecuredValidatorRequiredScope="requiredScope"')  # noqa: E501
                    jaas_blob.append(';')

            jaas_blob.append('};\n')
            self.conf['jaas_file'] = self.create_file('jaas_broker.conf',
                                                      data='\n'.
                                                      join(jaas_blob))
            self.env_add('KAFKA_OPTS',
                         '-Djava.security.auth.login.config=%s' %
                         self.conf['jaas_file'])
            if self.cluster.debug:
                self.env_add('KAFKA_OPTS', '-Djava.security.debug=all')

        # SSL config and keys (et.al.)
        if ssl is not None:
            keystore, truststore, _, _ = ssl.create_keystore('broker%s' %
                                                             self.appid)
            conf_blob.append('ssl.protocol=TLS')
            conf_blob.append('ssl.enabled.protocols=TLSv1.2,TLSv1.1,TLSv1')
            conf_blob.append('ssl.keystore.type = JKS')
            conf_blob.append('ssl.keystore.location = %s' % keystore)
            conf_blob.append('ssl.keystore.password = %s ' %
                             ssl.conf.get('ssl_key_pass'))
            conf_blob.append('ssl.key.password = %s' %
                             ssl.conf.get('ssl_key_pass'))
            conf_blob.append('ssl.truststore.type = JKS')
            conf_blob.append('ssl.truststore.location = %s' % truststore)
            conf_blob.append('ssl.truststore.password = %s' %
                             ssl.conf.get('ssl_key_pass'))
            conf_blob.append('ssl.client.auth = %s' %
                             self.conf.get('ssl_client_auth', 'required'))

        # Enable JMX
        jmx_port = trivup.TcpPortAllocator(self.cluster).next(self)
        # FIXME: JmxTool does not work when JMX is bound to listener_host,
        #        so we unfortunately can't bind to listener_host here.
        self.conf['jmx_port'] = jmx_port
        self.env_add('JMX_PORT', str(jmx_port))

        # Generate config file
        self.conf['conf_file'] = self.create_file_from_template('server.properties',  # noqa: E501
                                                                self.conf,
                                                                append_data=Template('\n'.join(conf_blob)).substitute(self.conf))  # noqa: E501

        # Generate LOG4J file (if app debug is enabled)
        if self.debug:
            self.conf['log4j_file'] = self.create_file_from_template('log4j.properties', self.conf, subst=False)  # noqa: E501
            self.env_add('KAFKA_LOG4J_OPTS',
                         '-Dlog4j.configuration=file:%s' %
                         self.conf['log4j_file'])

        self.env_add('LOG_DIR', self.mkpath('debug'))

        # Runs in foreground, stopped by Ctrl-C
        # This is the default for no-deploy use:
        # will be overwritten by deploy() if enabled.

        self.conf['start_cmd'] = '%s %s' % (start_sh, self.conf['conf_file'])
        self.conf['stop_cmd'] = None  # Ctrl-C

    def operational(self):
        self.dbg('Checking if operational')
        addr, port = self.get('address').split(':')
        with contextlib.closing(socket.socket(socket.AF_INET,
                                              socket.SOCK_STREAM)) as s:
            return s.connect_ex((addr, int(port))) == 0

    def kraft_setup_storage(self):
        """ Set up KRaft storage """
        cmd = '{}/kafka-storage.sh format -t {} -c {}'.format(
            self.conf['bindir'],
            self.cluster.uuid,
            self.conf['conf_file'])
        self.dbg('KRaft: setting up storage with: {}'.format(cmd))
        r = os.system(cmd)
        if r != 0:
            raise Exception('KRaft setup failed: "%s" returned exit code %d' %
                            (cmd, r))

    def kraft_configure_controllers(self):
        """ Configure the KRaft controllers.
        Needs to be performed in the deploy stage when all KafkaBrokerApps have
        been instantiated. """

        # Find all controllers (all KafkaBrokerApps for now) and construct
        # a list of broker_id@host:port.
        controllers = [
            '{}@{}'.format(
                x.appid, x.conf['controller_listener'].split('://')[-1])
            for x in self.cluster.find_apps(self.__class__)]

        self.dbg('KRaft: controllers: {}'.format(controllers))
        with open(self.conf['conf_file'], 'a') as f:
            f.write('controller.quorum.voters={}\n'.format(
                ','.join(controllers)))

    def kraft_setup(self):
        """ Set up KRaft. Should be called from deploy(). """
        self.kraft_configure_controllers()
        self.kraft_setup_storage()

    def deploy(self):
        destdir = os.path.join(self.cluster.mkpath(self.__class__.__name__),
                               'kafka', self.get('version'))
        self.dbg('Deploy %s version %s on %s to %s' %
                 (self.name, self.get('version'), self.node.name, destdir))
        deploy_exec = self.resource_path('deploy.sh')
        if not os.path.exists(deploy_exec):
            raise NotImplementedError('Kafka deploy.sh script missing in %s' %
                                      deploy_exec)
        t_start = time.time()
        cmd = '%s %s "%s" "%s"' % \
              (deploy_exec, self.get('version'),
               self.get('kafka_path', destdir), destdir)
        self.dbg('Deploy command: {}'.format(cmd))
        r = os.system(cmd)
        if r != 0:
            raise Exception('Deploy "%s" returned exit code %d' % (cmd, r))
        self.dbg('Deployed version %s in %ds' %
                 (self.get('version'), time.time() - t_start))

        self.conf['destdir'] = destdir
        self.conf['bindir'] = os.path.join(self.conf['destdir'], 'bin')

        if self.kraft:
            self.kraft_setup()

        # Override start command with updated path.
        self.conf['start_cmd'] = '%s/bin/kafka-server-start.sh %s' % \
                                 (destdir, self.conf['conf_file'])
        self.dbg('Updated start_cmd to %s' % self.conf['start_cmd'])
        # Add kafka-dir/bin to PATH so that the bundled tools are
        # easily called.
        self.env_add('PATH', os.environ.get('PATH') + ':' +
                     os.path.join(destdir, 'bin'), append=False)

    def _add_simple_authorizer(self, conf_blob):
        # Kafka removed SimpleAclAuthorizer class in v3.0.0
        # https://github.com/apache/kafka/commit/976e78e405d57943b989ac487b7f49119b0f4af4#diff-e0ccf1b5c964d2c303b6a69a8b8b67df5a6bfbae8aa514f580d353c4c6bf8e36
        if self.version[0] >= 3:
            conf_blob.append('authorizer.class.name=kafka.security.authorizer.AclAuthorizer')   # noqa: E501
        else:
            conf_blob.append('authorizer.class.name=kafka.security.auth.SimpleAclAuthorizer')   # noqa: E501



#!/usr/bin/env python
#

# Copyright (c) 2016-2019, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from trivup import trivup
from trivup.apps.KafkaBrokerApp import KafkaBrokerApp
import os


class ZookeeperApp (trivup.App):
    """ Zookeeper server app """

    def __init__(self, cluster, conf=None, on=None, bin_path=None):
        """
        @param cluster     Current cluster
        @param on          Node name to run on

        Config:
          bindir    Path to zookeeper-server-start.sh directory (optional)
                    Falls back to Kafka bindir
          zk_port   Port at which Zookeeper should be bound (optional)
                    A (random) free port will be chosen otherwise

        Exposes 'address' (host:port) for other apps.
        """
        super(ZookeeperApp, self).__init__(cluster, conf=conf, on=on)
        self.conf['port'] = trivup.TcpPortAllocator(self.cluster).next(
            self, port_base=self.conf.get('zk_port', None))
        self.conf['datadir'] = self.create_dir('datadir')
        self.conf['address'] = '%(nodename)s:%(port)d' % self.conf
        # Generate config file
        self.conf['conf_file'] = self.create_file_from_template('zookeeper.properties', self.conf)  # noqa: E501

    def start_cmd(self):
        bindir = self.get('bindir', None)
        if bindir is None:
            k = self.cluster.find_app(KafkaBrokerApp)
            if k is not None:
                bindir = k.get('bindir', None)

        zkbin = 'zookeeper-server-start.sh'
        if bindir:
            zkbin = os.path.join(bindir, zkbin)

        return '%s %s' % (zkbin, self.conf['conf_file'])

    def operational(self):
        self.dbg('Checking if operational')
        return os.system(('(echo srvr | nc %s | grep -q Zookeeper.version) '
                          '2>/dev/null') %
                         ' '.join(self.get('address').split(':'))) == 0

    def deploy(self):
        """ Deploy is a no-op for ZK since it is run from Kafka dir """
        pass



#!/usr/bin/env python
#

# Copyright (c) 2016-2019, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from trivup import trivup
import os


class KerberosKdcApp (trivup.App):
    """ Kerberos KDC app
        kdc must be installed on the target node.
    """
    def __init__(self, cluster, realm, conf=None, on=None):
        """
        @param cluster     Current cluster
        @param realm       Realm name
        @param conf        Configuration dict, optional.
           "port": port to bind to.
           "cross_realms": "realm1=kdc1:port,realm2=kdc2:port" - cross-realm.
                           The first cross realm is the $default_realm.
           "renew_lifetime": see krb5.conf docs  (default 10 min)
           "ticket_lifetime": see krb5.conf docs (default 60 min)
        @param on          Node name to run on
        """
        super(KerberosKdcApp, self).__init__(cluster, conf=conf, on=on)

        self.conf['realm'] = realm
        if self.conf.get('port', None) is None:
            self.conf['port'] = trivup.TcpPortAllocator(self.cluster).next(self)  # noqa: E501
        self.conf['address'] = '%(nodename)s:%(port)d' % self.conf
        self.conf['dbpath'] = self.mkpath('database')
        self.conf['admin_keytab'] = self.mkpath('admin_keytab')
        self.conf['stash_file'] = self.mkpath('stash_file')

        if self.conf.get('renew_lifetime', None) is None:
            self.conf['renew_lifetime'] = '12h'

        if self.conf.get('ticket_lifetime', None) is None:
            self.conf['ticket_lifetime'] = '30m'

        # Set up cross-realm trusts, if desired.
        cross_realms = self.conf.get('cross_realms', '').split(',')
        if len(cross_realms) > 0 and cross_realms[0] != '':
            cross_realms_conf = ""
            capaths_conf = ""
            for crinfo in cross_realms:
                crealm, ckdc = crinfo.split('=')
                if crealm == realm:
                    continue
                cross_realms_conf += " %s = {\n  kdc = %s\n  admin_server = %s\n }\n" % (crealm, ckdc, ckdc)  # noqa: E501
                capaths_conf += " %s = {\n  %s = .\n }\n" % (crealm, realm)
                capaths_conf += " %s = {\n  %s = .\n }\n" % (realm, crealm)

            self.conf['default_realm'] = cross_realms[0].split('=')[0]
            self.conf['cross_realms'] = cross_realms_conf
            self.conf['capaths'] = capaths_conf
        else:
            self.conf['default_realm'] = realm
            self.conf['cross_realms'] = ''
            self.conf['capaths'] = ''

        # Generate config files
        self.conf['krb5_conf'] = self.create_file_from_template('krb5.conf',
                                                                self.conf)
        self.env_add('KRB5_CONFIG', self.conf['krb5_conf'])
        self.conf['kdc_conf'] = self.create_file_from_template('kdc.conf',
                                                               self.conf)
        self.env_add('KRB5_KDC_PROFILE', self.conf['kdc_conf'])

        # Create database and stash file
        r = self.execute('kdb5_util -P "" -r %(realm)s -d "%(dbpath)s" -sf "%(stash_file)s" create -s' % self.conf).wait()  # noqa: E501
        if r != 0:
            raise Exception('Failed to create kdb5 database')

        self.conf['start_cmd'] = '/usr/sbin/krb5kdc -n'
        self.conf['stop_cmd'] = None  # Ctrl-C

    def operational(self):
        self.dbg('Checking if operational: FIXME')
        return True

    def deploy(self):
        """ Requires krb5kdc to be installed through other means, e.g.:
             sudo apt-get install krb5-kdc krb5-admin-server """
        pass

    def add_principal(self, primary, instance=None):
        """
        @brief Add principal to server and generate keytab
        @param primary The principal primary ("primary/instance@realm")
        @param instance The principal instance, optional.
        @returns (principal string, keytab path)
        """
        # Add principal
        if instance is not None:
            principal = '%s/%s@%s' % (primary, instance, self.conf['realm'])
        else:
            principal = '%s@%s' % (primary, self.conf['realm'])

        r = self.execute('kadmin.local -d "%s" -q "addprinc -randkey %s"' %
                         (self.conf.get('dbpath'), principal)).wait()
        if r != 0:
            raise Exception('ktadmin addprinc failed')

        # Generate keytab
        keytabdir = self.create_dir(os.path.join('keytabs', primary))
        if instance is not None:
            keytab = self.mkpath(os.path.join(keytabdir, instance))
        else:
            keytab = self.mkpath(os.path.join(keytabdir, "default"))

        r = self.execute('kadmin.local -d "%s" -q "ktadd -k "%s" %s"' %
                         (self.conf.get('dbpath'), keytab, principal)).wait()
        if r != 0:
            raise Exception('ktadmin ktadd failed')

        # Return keytab path
        return (principal, keytab)

    @staticmethod
    def add_cross_realm_tgts(kdcs):
        """ Add cross-realm TGTs.
            kdcs is a dict indexed by realm name, value is KerberosKdcApp. """
        realms = kdcs.keys()
        for realm in realms:
            for crealm in [x for x in realms if x != realm]:
                kdcs[realm].execute('kadmin.local -d "{}" -q "addprinc -requires_preauth -pw password krbtgt/{}@{}"'.format(kdcs[realm].conf.get('dbpath'), crealm, realm)).wait()  # noqa: E501
                kdcs[realm].execute('kadmin.local -d "{}" -q "addprinc -requires_preauth -pw password krbtgt/{}@{}"'.format(kdcs[realm].conf.get('dbpath'), realm, crealm)).wait()  # noqa: E501



from distutils.core import setup
from setuptools import find_packages
from glob import glob

data = dict()
app = 'trivup'
data[app] = list()
# Find Apps data
for d in glob('trivup/apps/*App'):
    data[app] += [x[x.find('apps/'):] for x in glob('%s/*' % d)
                  if x[-1:] != '~']

setup(name='trivup',
      version='0.12.0',
      description='Trivially Up a cluster of programs, such as a Kafka cluster',  # noqa: E501
      author='Magnus Edenhill',
      author_email='magnus@edenhill.se',
      url='https://github.com/edenhill/trivup',
      license_files=['LICENSE'],
      packages=find_packages(),
      package_data=data,
      install_requires=[
          'requests',
          'jwcrypto',
          'python_jwt'
      ],
      classifiers=[
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 2",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent",
      ])



#!/usr/bin/env python
#

# Copyright (c) 2016-2019, Magnus Edenhill
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from trivup import trivup

import os
import shutil
from copy import copy
from textwrap import dedent


class SslApp (trivup.App):
    """ Generates SSL certificates for use by other apps.
        This is not a running app but simply provides helper methods
        to generate certificates, etc. """

    def __init__(self, cluster, conf=None, on=None):
        """
        @param cluster     Current cluster
        @param conf        Configuration dict, see below.
        @param on          Node name to run on

        Honoured @param conf properties:
         * ssl_key_pass - SSL keytab password (default: 12345678)
         * SSL_{OU,O,L,S,ST,C} - (defaults: NN)

        """
        super(SslApp, self).__init__(cluster, conf=conf, on=on)

        self.conf.setdefault('ssl_key_pass', '12345678')
        self.conf.setdefault('ssl_OU', 'OU')
        self.conf.setdefault('ssl_O', 'O')
        self.conf.setdefault('ssl_L', 'L')
        self.conf.setdefault('ssl_ST', 'ST')
        self.conf.setdefault('ssl_S', 'S')
        self.conf.setdefault('ssl_C', 'NN')
        self.conf.setdefault('ssl_user', os.getenv('USER', 'NN'))

        # Generate two CA certs, the first one will be unused and the second
        # one will be what everything else is signed with.
        # This allows us to test multi-CA PEMs.
        self.unused_ca = self.create_ca_cert(self.__class__.__name__ + "_unused")
        self.ca = self.create_ca_cert(self.__class__.__name__)

        # Concatenate both PEMs to a single "all_cas" PEM file,
        # putting the unused CA first in the file.
        self.all_cas = {'pem': self.mkpath('all_cas.pem')}
        with open(self.all_cas['pem'], 'w') as f:
            for pemfile in [self.unused_ca['pem'], self.ca['pem']]:
                      with open(pemfile, 'r') as pf:
                            shutil.copyfileobj(pf, f)

    def exec_cmd(self, cmd):
        """ Run command with args, raise exception on failure. """
        r = self.execute(cmd, stdout_fd=os.devnull).wait()
        if r != 0:
            raise Exception('%s exited with status code %d' % (cmd, r))

    def mksubj(self, cn):
        """ Generate a -subj argument string """
        d = copy(self.conf)
        d['ssl_CN'] = cn
        return "/C=%(ssl_C)s/ST=%(ssl_ST)s/L=%(ssl_L)s/O=%(ssl_O)s/CN=%(ssl_CN)s" % d  # noqa: E501

    def create_ca_cert(self, cn):
        """
        Create CA cert
        @returns {'pem': .., 'der': .., 'key': .., 'srl': .., 'password': ..}
        """
        ret = {'key': self.mkpath('ca_%s.key' % cn),
               'srl': self.mkpath('ca_%s.srl' % cn),
               'pem': self.mkpath('ca_%s.pem' % cn),
               'der': self.mkpath('ca_%s.der' % cn),
               'password': self.conf.get('ssl_key_pass')}

        self.dbg('Generating CA cert for %s in %s' % (cn, ret['pem']))
        self.exec_cmd('openssl req -new -x509 -keyout "%s" -out "%s" -days 10000 -passin "pass:%s" -passout "pass:%s" -subj "%s"' %  # noqa: E501
                      (ret['key'], ret['pem'],
                       ret['password'], ret['password'],
                       self.mksubj(cn)))

        self.dbg('Convert CA PEM to DER')
        self.exec_cmd('openssl x509 -outform der -in "%s" -out "%s"' %
                      (ret['pem'], ret['der']))
        return ret

    def create_keystore(self, cn):
        """
        Create signed Java keystore for @param cn
        @returns (keystore, truststore, cert, signedcert)
        """
        keystore = self.mkpath('%s.keystore.jks' % cn)
        truststore = self.mkpath('%s.truststore.jks' % cn)
        cert = self.mkpath('%s.cert' % cn)
        signedcert = self.mkpath('%s.signedcert' % cn)

        d = copy(self.conf)
        d.update({'ssl_CN': cn})
        inblob = """%(ssl_CN)s
%(ssl_OU)s
%(ssl_O)s
%(ssl_L)s
%(ssl_S)s
%(ssl_C)s
yes""" % d

        self.dbg('Generating key for %s: %s' % (cn, keystore))
        self.exec_cmd('keytool -keyalg RSA -storepass "%s" -keypass "%s" -keystore "%s" -alias localhost -validity 10000 -genkey <<EOF\n%s\nEOF' %  # noqa: E501
                      (self.conf.get('ssl_key_pass'),
                       self.conf.get('ssl_key_pass'),
                       keystore, inblob))

        self.dbg('Adding truststore for %s: %s' % (cn, truststore))
        self.exec_cmd('keytool -storepass "%s" -keypass "%s" -keystore "%s" -alias CARoot -import -file "%s" <<EOF\nyes\nEOF' %  # noqa: E501
                      (self.conf.get('ssl_key_pass'),
                       self.conf.get('ssl_key_pass'),
                       truststore, self.ca['pem']))

        self.dbg('Export certificate for %s: %s' % (cn, cert))
        self.exec_cmd('keytool -storepass "%s" -keypass "%s" -keystore "%s" -alias localhost -certreq -file "%s"' %  # noqa: E501
                      (self.conf.get('ssl_key_pass'),
                       self.conf.get('ssl_key_pass'),
                       keystore, cert))

        self.dbg('Sign certificate for %s' % cn)
        self.exec_cmd('openssl x509 -req -CA "%s" -CAkey "%s" -in "%s" -out "%s" -days 10000 -CAcreateserial -passin "pass:%s"' %  # noqa: E501
                      (self.ca['pem'], self.ca['key'],
                       cert, signedcert,
                       self.conf.get('ssl_key_pass')))

        self.dbg('Import CA for %s' % cn)
        self.exec_cmd('keytool -storepass "%s" -keypass "%s" -keystore "%s" -alias CARoot -import -file "%s" <<EOF\nyes\nEOF' %  # noqa: E501
                      (self.conf.get('ssl_key_pass'),
                       self.conf.get('ssl_key_pass'),
                       keystore, self.ca['pem']))

        self.dbg('Import signed CA for %s' % cn)
        self.exec_cmd('keytool -storepass "%s" -keypass "%s" -keystore "%s" -alias localhost -import -file "%s"' %  # noqa: E501
                      (self.conf.get('ssl_key_pass'),
                       self.conf.get('ssl_key_pass'),
                       keystore, signedcert))

        return (keystore, truststore, cert, signedcert)

    def create_cert(self, cn, through_intermediate=False, with_ca=True):
        """
        Create certificate/keys, in multiple formats (PEM, DER, PKCS#12),
        for @param cn.
        This is typically used for clients.
        The PKCS contains private key, public key, and CA cert
        Pass in through_intermediate=True to sign a CA -> intermediate -> cert
        chain instead of just CA -> cert.
        Pass in with_ca=False to generate a self-signed certificate instead of
        one that chains back to the CA.
        @returns {'priv': {'pem': .., 'der': ..},
                  'pub': {'pem': .., 'der': ..},
                  'pkcs': '..',
                  'req': '..',
                  'password': '..'}
        """

        password = self.conf.get('ssl_key_pass')

        ret = {'priv': {'pem': self.mkpath('%s-priv.pem' % cn),
                        'der': self.mkpath('%s-priv.der' % cn)},
               'pub': {'pem': self.mkpath('%s-pub.pem' % cn),
                       'der': self.mkpath('%s-pub.der' % cn)},
               'pkcs': self.mkpath('%s.pfx' % cn),
               'req': self.mkpath('%s.req' % cn),
               'password': password}

        # Generate an intermediate cert, if this is required.
        if through_intermediate:
            ret.update(self._generate_intermediate(cn, with_ca=with_ca))

        self.dbg('Generating key for %s: %s' % (cn, ret['priv']['pem']))
        self.exec_cmd('openssl genpkey -algorithm RSA -pass "pass:%s" -out "%s"' %  # noqa: E501
                      (password, ret['priv']['pem']))

        self.dbg('Generating request for %s: %s' % (cn, ret['req']))
        self.exec_cmd('openssl req -passin "pass:%s" -passout "pass:%s" -key "%s" -new -out "%s" -subj "%s"' %  # noqa: E501
                      (password, password,
                       ret['priv']['pem'], ret['req'], self.mksubj(cn)))

        if through_intermediate:
            self.dbg('Signing key for %s with intermediate cert' % (cn))
            self.exec_cmd('openssl x509 -req -in "%s" -CA "%s" -CAkey "%s" -CAserial "%s" -out "%s"' %  # noqa: E501
                        (ret['req'], ret['intermediate_pub']['pem'], ret['intermediate_priv']['pem'],
                        self.ca['srl'], ret['pub']['pem']))
        elif with_ca:
            self.dbg('Signing key for %s with CA cert' % (cn))
            self.exec_cmd('openssl x509 -req -passin "pass:%s" -in "%s" -CA "%s" -CAkey "%s" -CAserial "%s" -out "%s"' %  # noqa: E501
                        (password,
                        ret['req'], self.ca['pem'], self.ca['key'],
                        self.ca['srl'], ret['pub']['pem']))
        else:
            self.dbg('Signing key for %s with self' % (cn))
            self.exec_cmd('openssl x509 -req -passin "pass:%s" -in "%s" -signkey "%s" -out "%s"' %  # noqa: E501
                        (password,
                        ret['req'], ret['priv']['pem'], ret['pub']['pem']))

        self.dbg('Converting public-key X.509 to DER for %s' % cn)
        self.exec_cmd('openssl x509 -outform der -in "%s" -out "%s"' %  # noqa: E501
                      (ret['pub']['pem'], ret['pub']['der']))

        self.dbg('Converting private-key X.509 to DER for %s' % cn)
        self.exec_cmd('openssl rsa -outform der -passin "pass:%s" -in "%s" -out "%s"' %  # noqa: E501
                      (password, ret['priv']['pem'], ret['priv']['der']))

        self._export_pkcs12(
            ret, cn, through_intermediate=through_intermediate, with_ca=with_ca)

        return ret

    def _generate_intermediate(self, cn, with_ca):
        ssl_cfg = self.mkpath('%s.cnf' % cn)
        password = self.conf.get('ssl_key_pass')
        ret = {
            'intermediate_priv': {'pem': self.mkpath('%s-intermediate-priv.pem' % cn),
                                  'der': self.mkpath('%s-intermediate-priv.der' % cn)},
            'intermediate_pub': {'pem': self.mkpath('%s-intermediate-pub.pem' % cn),
                                 'der': self.mkpath('%s-intermediate-pub.der' % cn)},
            'intermediate_req': self.mkpath('%s-intermediate.req' % cn),
        }

        with open(ssl_cfg, 'w') as f:
            f.write(dedent("""
                [req]
                distinguished_name=dn
                [ dn ]
                [ ext ]
                basicConstraints=CA:TRUE,pathlen:0
            """))

        self.dbg('Generating key for %s intermediate: %s' %
                 (cn, ret['intermediate_priv']['pem']))
        self.exec_cmd('openssl genpkey -algorithm RSA -out "%s"' %  # noqa: E501
                    (ret['intermediate_priv']['pem']))
        self.dbg('Generating request for %s: %s' %
                 (cn, ret['intermediate_req']))
        self.exec_cmd('openssl req -config "%s" -extensions ext -key "%s" -new -out "%s" -subj "%s"' %  # noqa: E501
                (ssl_cfg, ret['intermediate_priv']['pem'], ret['intermediate_req'], self.mksubj('%s-intermediate' % (cn))))

        # Work out if "intermediate" cert should be self-signed or actually signed by the CA.
        if with_ca:
            self.dbg('Signing key for %s intermediate with CA cert' % (cn))
            self.exec_cmd('openssl x509 -req -extfile "%s" -extensions ext -passin "pass:%s" -in "%s" -CA "%s" -CAkey "%s" -CAserial "%s" -out "%s"' %  # noqa: E501
                        (ssl_cfg, password,
                        ret['intermediate_req'], self.ca['pem'], self.ca['key'],
                        self.ca['srl'], ret['intermediate_pub']['pem']))
        else:
            self.dbg('Signing key for %s intermediate with self' % (cn))
            self.exec_cmd('openssl x509 -req -passin "pass:%s" -in "%s" -signkey "%s" -out "%s"' %  # noqa: E501
                        (password,
                        ret['intermediate_req'], ret['intermediate_priv']['pem'], ret['intermediate_pub']['pem']))

        self.dbg('Converting public-key X.509 to DER for %s intermediate' % cn)
        self.exec_cmd('openssl x509 -outform der -in "%s" -out "%s"' %  # noqa: E501
                    (ret['intermediate_pub']['pem'], ret['intermediate_pub']['der']))

        self.dbg('Converting private-key X.509 to DER for %s intermediate' % cn)
        self.exec_cmd('openssl rsa -outform der -passin "pass:%s" -in "%s" -out "%s"' %  # noqa: E501
                    (password, ret['intermediate_priv']['pem'], ret['intermediate_priv']['der']))

        return ret

    def _export_pkcs12(self, ret, cn, through_intermediate, with_ca):
        password = self.conf.get('ssl_key_pass')
        additional_certs_for_pkcs12 = []
        if through_intermediate:
            additional_certs_for_pkcs12.append(ret['intermediate_pub']['pem'])
        if with_ca:
            additional_certs_for_pkcs12.append(self.unused_ca['pem'])
            additional_certs_for_pkcs12.append(self.ca['pem'])
        certfile_arguments = ' '.join(
            ['-certfile "{}"'.format(c) for c in additional_certs_for_pkcs12])

        self.dbg('Creating PKCS#12 for %s in %s' % (cn, ret['pkcs']))
        self.exec_cmd('openssl pkcs12 -export -descert -out "%s" -inkey "%s" -in "%s"  %s -passin "pass:%s" -passout "pass:%s"' %  # noqa: E501
                      (ret['pkcs'],
                       ret['priv']['pem'],
                       ret['pub']['pem'],
                       certfile_arguments,
                       password, password))

    def operational(self):
        return True

    def deploy(self):
        return

    def start_cmd(self):
        return None
